#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#define MY_DEVICE_NAME "my_device"
#define MY_CLASS_NAME "my_class"

#define WR_VALUE _IOW('a', 'a', int32_t *)
#define RD_VALUE _IOR('a', 'b', int32_t *)

int32_t val = 0;
dev_t dev = 0;
static struct class *dev_class;
static struct cdev my_cdev;

static int my_open(struct inode *inode, struct file *file)
{
  printk(KERN_INFO "my_open()\n");
  return 0;
}

static int my_release(struct inode *inode, struct file *file)
{
  printk(KERN_INFO "my_release()\n");
  return 0;
}

static ssize_t my_read(struct file *file, char __user *buf, size_t len,
                       loff_t *off)
{
  ssize_t ret;
  printk(KERN_INFO "my_read()\n");

  if ((ret = copy_to_user((int32_t *) buf, &val, len))) {
    printk(KERN_ERR "Error: copy_to_user()\n");
    return ret;
  }

  return 0;
}

static ssize_t my_write(struct file *file, const char __user *buf, size_t len,
                        loff_t *off)
{
  ssize_t ret;
  printk(KERN_INFO "my_write()\n");

  if ((ret = copy_from_user(&val, (int32_t *) buf, len))) {
    printk(KERN_ERR "Error: copy_from_user()\n");
    return ret;
  }

  printk(KERN_INFO "val = %d\n", val);
  return len;
}

static long my_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
  printk(KERN_INFO "my_ioctl()\n");

  switch (cmd) {
  case WR_VALUE:
    my_write(file, (const char *) arg, sizeof(val), 0);
    break;

  case RD_VALUE:
    my_read(file, (char *) arg, sizeof(val), 0);
    break;

  default:
    printk(KERN_ERR "Error: unknown cmd = %u\n", cmd);
    break;
  }

  return 0;
}

static struct file_operations fops = {
  .owner          = THIS_MODULE,
  .read           = my_read,
  .write          = my_write,
  .open           = my_open,
  .unlocked_ioctl = my_ioctl,
  .release        = my_release,
};

static int __init my_driver_init(void)
{
  printk("my_driver_init()\n");

  if (alloc_chrdev_region(&dev, 0, 1, MY_DEVICE_NAME) < 0) {
    printk(KERN_ERR "Error: cannot allocate major number\n");
    return -1;
  }

  printk(KERN_INFO "major = %d, minor = %d\n", MAJOR(dev), MINOR(dev));

  cdev_init(&my_cdev, &fops);

  if (cdev_add(&my_cdev, dev, 1) < 0) {
    printk(KERN_ERR "Error: cannot add the device to the system\n");
    goto error_class;
  }

  if (IS_ERR(dev_class = class_create(THIS_MODULE, MY_CLASS_NAME))) {
    printk(KERN_ERR "Error: cannot create the struct class\n");
    goto error_class;
  }

  if (IS_ERR(device_create(dev_class, NULL, dev, NULL, MY_DEVICE_NAME))) {
    printk(KERN_ERR "Error: cannot create %s\n", MY_DEVICE_NAME);
    goto error_device;
  }

  return 0;

error_device:
  class_destroy(dev_class);
error_class:
  unregister_chrdev_region(dev, 1);
  return -1;
}

static void __exit my_driver_exit(void)
{
  printk("my_driver_exit()\n");
  device_destroy(dev_class, dev);
  class_destroy(dev_class);
  cdev_del(&my_cdev);
  unregister_chrdev_region(dev, 1);
}

module_init(my_driver_init);
module_exit(my_driver_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Hiroyuki Chishiro");
MODULE_DESCRIPTION("IOCTL Kernel Module");
